/**
 * Perfhawk InfluxDB Backend Listener for Loadrunner
 * by PE Enablement
 *
 * Based on: https://github.com/nigelargriffiths/InfluxDB-C-client
 */


#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <math.h>
#include <string.h>


#define DEBUG   if(debug)
#define MEGABYTE ( 1024 * 1024 ) /* USed as the default buffer sizes */

int debug = 0; /* 0=off, 1=on basic, 2=trace like output */

//char influx_hostname[1024 + 1] = { 0 };/* details of the influxdb server or telegraf */
char influx_ip[16 + 1] = { 0 };
long influx_port = 0;

char influx_database[256+1];		/* the influxdb database  */
char influx_username[64+1];		/* optional for influxdb access */
char influx_password[64+1];		/* optional for influxdb access */

char *output; /* all the stats must fit in this buffer */
long output_size = 0;
long output_char = 0;

char *influx_tags; /* saved tags for every influxdb line protocol mesurement */

int subended = 0;		/* stop ic_subend and ic_measureend both enig the measure */
int first_sub = 0;		/* need to remove the ic_measure measure before adding ic_sub measure */
char saved_section[64];
char saved_sub[64];

//int sockfd;			/* file desciptor for socket connection */

void error(char *buf)
{
	lr_error_message("error: \"%s\"", buf);
    //close(sockfd);
    //sleep(2);			/* this can help the socket close cleanly at the remote end */
    //exit(1);
    lr_exit(LR_EXIT_VUSER, LR_FAIL);
}

void ic_debug(int level)
{
	debug = level;
}

/* ic_tags() argument is the measurement tags for influddb */
/* example: "host=vm1234"   note:the comma & hostname of the virtual machine sending the data */
/* complex: "host=lpar42,serialnum=987654,arch=power9" note:the comma separated list */
void ic_tags(char *t)
{
    lr_debug_message(LR_MSG_CLASS_RESULT_DATA | LR_MSG_CLASS_PARAMETERS, "ic_tags(%s)\n",t);
    if( influx_tags == (char *) 0) {
        if( (influx_tags = (char *)malloc(MEGABYTE)) == (char *)-1)
           error("failed to malloc() tags buffer");
    }

    strncpy(influx_tags,t,256);
}

void ic_influx_database(char *host, long port, char *db) /* note: converts influxdb hostname to ip address */
{
	//struct hostent *he;
	//char errorbuf[1024 +1 ];
	//char* h_addr_list;

	influx_port = port;
	strncpy(influx_database,db,256);

	// If host is an ip
	if(host[0] >= '0' && host[0] <='9') {
		lr_debug_message(LR_MSG_CLASS_RESULT_DATA | LR_MSG_CLASS_PARAMETERS,"ic_influx(ipaddr=%s,port=%ld,database=%s))\n",host,port,db);
		strncpy(influx_ip,host,16);
	}
	// If host is a domain name
	else {
		error("Must use ip");
		return;
    }
}

void ic_influx_userpw(char *user, char *pw)
{
	lr_debug_message(LR_MSG_CLASS_RESULT_DATA | LR_MSG_CLASS_PARAMETERS,"ic_influx_userpw(username=%s,pssword=%s))\n",user,pw);
	strncpy(influx_username,user,64);
	strncpy(influx_password,pw,64);
}

void ic_check(long adding) /* Check the buffer space */
{
    if(output == (char *)0) {			/* First time create the buffer *
	if( (output = (char *)malloc(MEGABYTE)) == (char *)-1)
	    error("failed to malloc() output buffer");
    }
    if(output_char + (2*adding) > output_size) /* When near the end of the output buffer, extend it*/
	if( (output = (char *)realloc(output, output_size + MEGABYTE)) == (char *)-1)
	    error("failed to realloc() output buffer");
    }
}

void remove_ending_comma_if_any()
{
    if (output[output_char - 1] == ',') {
	output[output_char - 1] = 0;	/* remove the char */
	output_char--;
    }
}

void ic_measure(char *section)
{
    ic_check( strlen(section) + strlen(influx_tags) + 3);

    output_char += sprintf(&output[output_char], "%s,%s ", section, influx_tags);
    strcpy(saved_section, section);
    first_sub = 1;
    subended = 0;
    lr_debug_message(LR_MSG_CLASS_RESULT_DATA | LR_MSG_CLASS_PARAMETERS,"ic_measure(\"%s\") count=%ld\n", section, output_char);
}

void ic_measureend()
{
    ic_check( 4 );
    remove_ending_comma_if_any();
    if (!subended) {
         output_char += sprintf(&output[output_char], "   \n");
    }
    subended = 0;
    lr_debug_message(LR_MSG_CLASS_RESULT_DATA | LR_MSG_CLASS_PARAMETERS,"ic_measureend()\n");
}

void ic_timestamp()
{
	typedef long time_t;
	time_t t;
	int rightNow;

	rightNow = time(&t);

	ic_check(10);
    remove_ending_comma_if_any();

	output_char += sprintf(&output[output_char], " %d",rightNow);

	lr_debug_message(LR_MSG_CLASS_RESULT_DATA | LR_MSG_CLASS_PARAMETERS,"ic_timestamp(time=%d)\n",rightNow);
}

/* Note this added a further tag to the measurement of the "resource_name" */
/* measurement might be "disks" */
/* sub might be "sda1", "sdb1", etc */
void ic_sub(char *resource)
{
    int i;

    ic_check( strlen(saved_section) + strlen(influx_tags) +strlen(saved_sub) + strlen(resource) + 9);

    /* remove previously added section */
    if (first_sub) {
	for (i = output_char - 1; i > 0; i--) {
	    if (output[i] == '\n') {
		output[i + 1] = 0;
		output_char = i + 1;
		break;
	    }
	}
    }
    first_sub = 0;

    /* remove the trailing s */
    strcpy(saved_sub, saved_section);
    if (saved_sub[strlen(saved_sub) - 1] == 's') {
	saved_sub[strlen(saved_sub) - 1] = 0;
    }
    output_char += sprintf(&output[output_char], "%s,%s,%s_name=%s ", saved_section, influx_tags, saved_sub, resource);
    subended = 0;
    lr_debug_message(LR_MSG_CLASS_RESULT_DATA | LR_MSG_CLASS_PARAMETERS,"ic_sub(\"%s\") count=%ld\n", resource, output_char);
}

void ic_subend()
{
    ic_check( 4 );
    remove_ending_comma_if_any();
    output_char += sprintf(&output[output_char], "   \n");
    subended = 1;
    lr_debug_message(LR_MSG_CLASS_RESULT_DATA | LR_MSG_CLASS_PARAMETERS,"ic_subend()\n");
}

void ic_long(char *name, long long value)
{
    ic_check( strlen(name) + 16 + 4 );
    output_char += sprintf(&output[output_char], "%s=%lldi,", name, value);
    lr_debug_message(LR_MSG_CLASS_RESULT_DATA | LR_MSG_CLASS_PARAMETERS,"ic_long(\"%s\",%lld) count=%ld\n", name, value, output_char);
}

void ic_double(char *name, double value)
{
    ic_check( strlen(name) + 16 + 4 );
    if (isnan(value) ) { /* not-a-number or infinity */
		lr_debug_message(LR_MSG_CLASS_RESULT_DATA | LR_MSG_CLASS_PARAMETERS,"ic_double(%s,%.1f) - nan error\n", name, value);
    } else {
		output_char += sprintf(&output[output_char], "%s=%.3f,", name, value);
		lr_debug_message(LR_MSG_CLASS_RESULT_DATA | LR_MSG_CLASS_PARAMETERS,"ic_double(\"%s\",%.1f) count=%ld\n", name, value, output_char);
    }
}

void ic_string(char *name, char *value)
{
    int i;
    int len;

    ic_check( strlen(name) + strlen(value) + 4 );
    len = strlen(value);
    for (i = 0; i < len; i++) 	/* replace problem characters and with a space */
	if (value[i] == '\n' || iscntrl(value[i]))
	    value[i] = ' ';
    output_char += sprintf(&output[output_char], "%s=\"%s\",", name, value);
    lr_debug_message(LR_MSG_CLASS_RESULT_DATA | LR_MSG_CLASS_PARAMETERS,"ic_string(\"%s\",\"%s\") count=%ld\n", name, value, output_char);
}

void ic_push()
{
	char url[1024];
    char header[1024];
    char result[1024];
    char buffer[1024 * 8];
    int ret;
    int i;
    int total;
    int sent;
    int code;

    if (output_char == 0)	/* nothing to send so skip this operation */
		return;

	if (influx_port == 0) {
    		error("influx port is not set, bailing out");
    		return;
	}

	lr_debug_message(LR_MSG_CLASS_RESULT_DATA | LR_MSG_CLASS_PARAMETERS,"ic_push() size=%ld\n", output_char);
	lr_debug_message(LR_MSG_CLASS_RESULT_DATA | LR_MSG_CLASS_PARAMETERS,"ic_push() output=%s\n", output);

	sprintf(url,"URL=http://%s:%ld/write?db=%s&u=%s&p=%s&precision=s",influx_ip, influx_port, influx_database, influx_username, influx_password);
	lr_debug_message(LR_MSG_CLASS_RESULT_DATA | LR_MSG_CLASS_PARAMETERS,"ic_push() url=%s\n", url);

	sprintf(buffer,"Body=%s",output);

	//lr_continue_on_error(1);
	web_custom_request("InfluxWriter","Method=POST",url,buffer,LAST);
	//lr_continue_on_error(0);

    output[0] = 0;
    output_char = 0;
}


/////////////////////////////////////////////////////////////

//49.27.22.243. pcoe01/1111


int ph_hasInit = 0;
int ph_lastPush = 0;
int PH_PUSH_THRESHOLD = 30;

int ph_start_transaction( char *transaction_name ) {
	if(ph_hasInit == 0) {
		ph_hasInit = 1;
		//ic_influx_database("49.27.22.243", 8086, "loadrunner");
		//ic_influx_userpw("pcoe01", "1111");
		ic_influx_database("10.209.1.131", 80, "loadrunner");
		ic_influx_userpw("pcoe01", "1111");
	}

	return lr_start_transaction(transaction_name);
}

int ph_end_transaction( char *transaction_name, int status) {
	int vuserid, scid, tagLength;
	char* tagValue;
	char* vuser_group;
	char* hostname;
	double duration;

	lr_stop_transaction(transaction_name);

	duration = lr_get_transaction_duration(transaction_name);

	lr_whoami(&vuserid, &vuser_group, &scid);

	hostname = (char*)lr_get_host_name(),

	tagLength = 34 + strlen(transaction_name) + 10 + strlen(hostname) + 10;
	tagValue = (char*) malloc( sizeof(char) * tagLength );

	sprintf(tagValue,"lg=%s,vuser=%d,transaction=%s,scenario=%d",
	        hostname,
	        vuserid,
	        transaction_name,
	        scid);
	lr_debug_message(LR_MSG_CLASS_RESULT_DATA | LR_MSG_CLASS_PARAMETERS,"tags: %s - length=%d tagLength=%d",tagValue,strlen(tagValue),tagLength);

	ic_tags(tagValue);
	ic_measure("transactions");
	ic_double("duration",duration);
	ic_long("status",status);
	ic_timestamp();
	ic_measureend();

	ph_flush(0);

	free(tagValue);

	return lr_end_transaction(transaction_name, status);
}

void ph_flush(int flushRightAway) {
	typedef long time_t;
	time_t t;
	int rightNow;

	rightNow = time(&t);

	if(flushRightAway == 1 || rightNow - ph_lastPush > PH_PUSH_THRESHOLD ) {
		ic_push();
		ph_lastPush = rightNow;
	}
}
